/* gtk-3dcanvas-glcamera - an OpenGl "view" of a gtk-3dcanvas */

/* Copyright (C) 2002, Martin Quinson */

/* This program is free software; you can redistribute it and/or modify it
   under the terms of the license (GNU LGPL) which comes with this package. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <gtk/gtkgl.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "gtk-3dcanvas.h" /* header of this module */
#include "gtk-3dcanvas-private.h"

/* Useful */
#define PI                      3.14159265358979323846264338327
#define DEG(angleRad)           ((angleRad) * 180 / PI)
#define RAD(angle360)           ((angle360) * PI / 180)
#define SQR(x)                  ((x) * (x))
#define SIGN(x)                 (((x) >= 0) ? 1.0 : -1.0)

/*-----------------\
| Class definition |
\-----------------*/
static void gtk_3dcanvas_cameragl_class_init (Gtk3dcanvasCameraGlClass *class);
static void gtk_3dcanvas_cameragl_init(Gtk3dcanvasCameraGl *cam);

static GObject *
gtk_3dcanvas_cameragl_constructor (GType                  type,
		 	   guint                  n_construct_properties,
			   GObjectConstructParam *construct_properties);
static void
gtk_3dcanvas_cameragl_finalize (GObject *object);

static gpointer parent_class = NULL;

/*-------------------\
| Internal functions |
\-------------------*/
static void set_cursor_glyph( int glyph, GdkWindow *window );

#if 0 /* FIXME: make a property for it */
static float camera_calc_fov( float lens_length );
static float camera_calc_lens_length( float fov );
#endif

static void camera_calc_xyz_pos( Gtk3dcanvasCameraGl *cam );
static void camera_calc_xyz_target( Gtk3dcanvasCameraGl *cam );
#if 0 /* FIXME: dead code */
static void camera_make_target( Gtk3dcanvasCameraGl *cam );
#endif

static void cameragl_redraw_all(Gtk3dcanvasCameraGlClass *klass);

void gtk_3dcanvas_cameragl_do_draw( Gtk3dcanvasCameraGl *cam);
/* once every initialization is done, draw recursively the root group */
static void draw_group(Gtk3dcanvasCameraGl *cam,
		       Gtk3dcanvasGroup *group);
static void draw_item(Gtk3dcanvasCameraGl *cam,
		      Gtk3dcanvasItem *item);
/* draw the axis and floating grid */
static void draw_aux_axes(Gtk3dcanvasCameraGl *cam);
static void draw_aux_grid(Gtk3dcanvasCameraGl *cam);

/*----[ gtk_3dcanvas_cameragl_get_type]---------------------------------*/
GType gtk_3dcanvas_cameragl_get_type () {
  static GType type = 0;
  
  if (!type) {
    static const GTypeInfo type_info = {
      sizeof(Gtk3dcanvasCameraGlClass),
      (GBaseInitFunc) NULL,
      (GBaseFinalizeFunc) NULL,
      (GClassInitFunc) gtk_3dcanvas_cameragl_class_init,
      (GClassFinalizeFunc) NULL,
      NULL,                   /* class_data */
      sizeof (Gtk3dcanvasCameraGl),
      0,                      /* n_preallocs */
      (GInstanceInitFunc) gtk_3dcanvas_cameragl_init,
    };

    type = g_type_register_static (gtk_drawing_area_get_type (),
				   "Gtk3dcanvasCameraGl",
				   &type_info, 0);
  }
  
  return type;
}
  

/*----[ gtk_3dcanvas_cameragl_class_init ]---------------------------------*/
static void gtk_3dcanvas_cameragl_class_init (Gtk3dcanvasCameraGlClass 
					      *klass) {
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  /* nothing to init in the class for now */
  object_class->constructor  = gtk_3dcanvas_cameragl_constructor;
  object_class->finalize     = gtk_3dcanvas_cameragl_finalize;

  klass->camera=g_array_new(FALSE,FALSE,
			    sizeof(Gtk3dcanvasCameraGl*));
  klass->idle_func_id=gtk_idle_add_priority(GDK_PRIORITY_REDRAW,
				    (GtkFunction) cameragl_redraw_all,
                                    object_class);
  /*FIXME: should call gtk_idle_remove(object_class->idle_func_id) when
    destroyed*/
  /*
  g_object_class_install_property (object_class,
                                   PROP_ATTRIB_LIST,
                                   g_param_spec_pointer ("canvas",
							 _("Attrib list"),
							 _("Pointer to the OpenGL configuration attribute list."),
							 G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
  */
}

static void cameragl_redraw_all(Gtk3dcanvasCameraGlClass *klass) {
  int i;

  for (i=0; i<klass->camera->len; i++) {
    gtk_3dcanvas_cameragl_do_draw(g_array_index(klass->camera,
						Gtk3dcanvasCameraGl*,
						i));
  }
}

/*----[ gtk_3dcanvas_cameragl_constructor ]-------------------------------*/
static GObject *
gtk_3dcanvas_cameragl_constructor (GType                  type,
			   guint                  n_construct_properties,
			   GObjectConstructParam *construct_properties)
{
  GObject *object;

  object = G_OBJECT_CLASS (parent_class)->constructor (type,
                                                       n_construct_properties,
                                                       construct_properties);

  return object;
}

/*----[ gtk_3dcanvas_cameragl_finalize ]-----------------------------------*/
static void
gtk_3dcanvas_cameragl_finalize (GObject *object)
{
  Gtk3dcanvasCameraGl *cam = GTK_3DCANVAS_CAMERAGL (object);
  Gtk3dcanvas *world;
  int pos; /* position of this camera in canvas's table */

  gtk_widget_destroy (cam->ogl_w);
  gtk_widget_destroy (cam->window_w);

  world = cam->canvas;

  for (pos=0; 
       pos<world->camera_count && cam !=world->camera[pos];
       pos++);
  if (pos == world->camera_count) {
    printf("gtk_3dcanvas_cameragl_finalize: Damnit, I was asked to remove a camera from the canvas, but it is not regitered\n");
    G_OBJECT_CLASS (parent_class)->finalize (object);
    return;
  }

  world->camera_count--;

  if (pos != world->camera_count) /* Not last camera? */
    memmove( &(world->camera[pos]), &(world->camera[pos + 1]),
	     (world->camera_count - pos) * sizeof(Gtk3dcanvasCameraGl *) );
  world->camera = g_realloc( world->camera, 
			     world->camera_count * sizeof(Gtk3dcanvasCameraGl *) );
  
  G_OBJECT_CLASS (parent_class)->finalize (object);
}

/*----[ gtk_3dcanvas_cameragl_init ]---------------------------------------*/
static void gtk_3dcanvas_cameragl_init(Gtk3dcanvasCameraGl *cam) {
} 

/* Change cursor appearance */
static void set_cursor_glyph( int glyph, GdkWindow *window ) {
  static GdkCursor *cursor;
  static int prev_glyph = -1;

  if (prev_glyph != -1)
    gdk_cursor_unref( cursor );
  cursor = gdk_cursor_new( glyph );
  gdk_window_set_cursor( window, cursor );
  prev_glyph = glyph;
}

/* The default function used to initialize the oGL rendering area */
static void default_init_function() {
  GLfloat light0_pos[4]   = { -50.0, 50.0, 0.0, 0.0 };
  GLfloat light0_color[4] = { .6, .6, .6, 1.0 }; /* white light */
  GLfloat light1_pos[4]   = {  50.0, 50.0, 0.0, 0.0 };
  GLfloat light1_color[4] = { .4, .4, 1, 1.0 };  /* cold blue light */
  
  /* remove back faces */
  glDisable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  
  /* speedups */
  glEnable(GL_DITHER);
  glShadeModel(GL_SMOOTH);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);
  
  /* light */
  glLightfv(GL_LIGHT0, GL_POSITION, light0_pos);
  glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0_color);
  glLightfv(GL_LIGHT1, GL_POSITION, light1_pos);
  glLightfv(GL_LIGHT1, GL_DIFFUSE,  light1_color);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  glEnable(GL_LIGHTING);
  
  glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);
}

/*-----------------\
| Public functions |
\-----------------*/
/**
 * gtk_3dcanvas_cameragl_new:
 * @canvas: From which canvas we want a new camera 
 * @glconfig: The configuration with which we should create the GL-enabled 
 *            drawing area
 * @initfunc: pointer to the function to use to init the GL rendering. 
 *            When NULL, a default one is used.
 *
 * Add a new camera to the given canvas, and returns it.
 *
 * Return value: the newly created camera 
 **/
GtkWidget* gtk_3dcanvas_cameragl_new(Gtk3dcanvas *canvas,
				       GdkGLConfig *glconfig,
				       void (*initfunc)(void)) {
  GObject *ret;
  Gtk3dcanvasCameraGl *camera;
  int new_cam_id;

  /* create the object */
  ret = g_object_new(GTK_3DCANVAS_CAMERAGL_TYPE,NULL); 

  /* Connect to the canvas */
  camera = GTK_3DCANVAS_CAMERAGL(ret);
  camera->canvas = canvas;

  new_cam_id = canvas->camera_count;
  canvas->camera_count++;  
  canvas->camera = g_realloc( canvas->camera, 
			      canvas->camera_count * 
			      sizeof(Gtk3dcanvasCameraGl *) );
  canvas->camera[new_cam_id] = camera;

  /* fill the camera */
  if (new_cam_id > 0) {
    /* Initialize camera with primary's state */
    Gtk3dcanvasCameraGl *first=canvas->camera[0];
    camera->mouse_sens = first->mouse_sens;
    camera->phi        = first->phi;
    camera->theta      = first->theta;
    camera->distance   = first->distance;
    memcpy(&(camera->target),&(first->target),sizeof(Gtk3dPoint));
    memcpy(&(camera->pos),   &(first->pos),   sizeof(Gtk3dPoint));
    camera->fov        = first->fov;
    camera->width      = first->width;
    camera->height     = first->height;
    camera->near_clip  = first->near_clip;
    camera->far_clip   = first->far_clip;
    memcpy(&(camera->background),&(first->background),sizeof(Gtk3dRgbColor));
    camera->axes       = first->axes;
    camera->grid       = first->grid;
    gtk_3dcanvas_cameragl_set_dgamma_correct(camera,first->dgamma_correct);
  } else {
    /* This IS the primary */
    gtk_3dcanvas_cameragl_reset( camera );
  }

  /* gl-enable the drawing area */
  gtk_widget_set_gl_capability (GTK_WIDGET (camera),
                                glconfig,
                                GDK_GL_RGBA_TYPE,
                                NULL,
                                TRUE);

  /* Signaling stuff */
  gtk_widget_set_events( GTK_WIDGET(camera), GDK_EXPOSURE_MASK |
			 GDK_BUTTON_MOTION_MASK | GDK_BUTTON1_MOTION_MASK |
			 GDK_BUTTON2_MOTION_MASK | GDK_BUTTON3_MOTION_MASK |
			 GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK );
  /* Attach the signal handlers */
  g_signal_connect( GTK_OBJECT(camera), "realize",
		    GTK_SIGNAL_FUNC(initfunc ? initfunc : default_init_function), NULL );
  
  g_signal_connect( GTK_OBJECT(camera), "expose_event",
		    GTK_SIGNAL_FUNC(gtk_3dcanvas_cameragl_expose), 
		    NULL );
  g_signal_connect( GTK_OBJECT(camera), "configure_event",
		    GTK_SIGNAL_FUNC(gtk_3dcanvas_cameragl_configure),
		    NULL );

  g_signal_connect( GTK_OBJECT(camera), "button_press_event",
		    GTK_SIGNAL_FUNC(gtk_3dcanvas_cameragl_mouse_move),
		    NULL );
  g_signal_connect( GTK_OBJECT(camera), "button_release_event",
		    GTK_SIGNAL_FUNC(gtk_3dcanvas_cameragl_mouse_move),
		    NULL );
  g_signal_connect( GTK_OBJECT(camera), "motion_notify_event",
		    GTK_SIGNAL_FUNC(gtk_3dcanvas_cameragl_mouse_move), 
		    NULL );

  /* Widget pointers will be assigned elsewhere
   * (in main_window( ) for primary, in menu_Camera_Spawn( ) for others) */
  camera->window_w = NULL;
  camera->ogl_w = NULL;

  g_array_append_val(GTK_3DCANVAS_CAMERAGL_GET_CLASS(camera)->camera,
		     camera);
  gtk_3dcanvas_cameragl_draw(camera);
  return GTK_WIDGET ( ret );
}


/** 
 * gtk_3dcanvas_cameragl_reset:
 * @cam: The camera to change
 *
 * Reset the camera to the default position/direction
 **/
void gtk_3dcanvas_cameragl_reset( Gtk3dcanvasCameraGl *cam ) {

  /* Convert user phi/theta to struct camera phi/theta values */
  cam->mouse_sens = cam->canvas->default_mouse_sens;
  cam->phi = fmod( cam->canvas->default_phi + 180.0, 360.0) ;
  cam->theta = - cam->canvas->default_theta;
  cam->distance = 3 * cam->canvas->root->extents.avg;
  cam->target.x = 0;
  cam->target.y = 0;
  cam->target.z = 0;
  cam->fov = cam->canvas->default_fov;
  cam->near_clip =  cam->canvas->root->extents.avg / 20;
  cam->far_clip =  cam->canvas->root->extents.avg * 50;
  camera_calc_xyz_pos( cam );
  cam->background.r = 1; /*FIXME: get default from canvas */
  cam->background.g = 1;
  cam->background.b = 1;
  gtk_3dcanvas_cameragl_set_dgamma_correct(cam,1);/*FIXME:default in canvas*/

  cam->axes = 0.0;
  cam->grid = 1.0;
  /* Update Camera Position dialog if it is active 
     FIXME: implement it/
  dialog_Camera_Position( NULL, MESG_(RESET) );
  */
  gtk_3dcanvas_cameragl_draw(cam);
}

#if 0 /* FIXME: dead code */
/**
 * gtk_3dcanvas_cameragl_of_widget:
 * @widget: widget to search the camera for
 * @canvas: where to search
 * 
 * Returns the camera associated to a (GL) widget
 **/
Gtk3dcanvasCameraGl *gtk_3dcanvas_cameragl_of_widget(GtkWidget   *widget,
							Gtk3dcanvas *canvas){
  Gtk3dcanvasCameraGl *cam;
  int i;
  
  for (i = 0; i < canvas->camera_count; i++) {
    cam = canvas->camera[i];
    if ((widget == cam->window_w) || (widget == cam->ogl_w))
      return cam;
  }
  
  return NULL; /* No camera associated with this widget */
}

/* Obtain a field of view (in degrees) from a lens length (in mm) */
float camera_calc_fov( float lens_length ) {
  return DEG(2 * atan( 21.6663 / (float)lens_length ));
}

/* Obtain a lens length (in mm) from a field of view (in degrees) */
float camera_calc_lens_length( float fov ) {
  return 21.6663 / tan( RAD(fov) / 2 );
}
#endif /* dead code */

/* Calculates a camera's XYZ target, given that the position 
 * and phi + theta + distance are defined */
static void camera_calc_xyz_pos( Gtk3dcanvasCameraGl *cam ) {
  float sin_phi, cos_phi, sin_theta, cos_theta;
  float dx, dy, dz;
  
  sin_phi = sin( RAD(cam->phi) );
  cos_phi = cos( RAD(cam->phi) );
  sin_theta = sin( RAD(cam->theta) );
  cos_theta = cos( RAD(cam->theta) );
  
  dx = cam->distance * cos_theta * cos_phi;
  dy = cam->distance * cos_theta * sin_phi;
  dz = cam->distance * sin_theta;
  
  cam->pos.x = cam->target.x + dx;
  cam->pos.y = cam->target.y + dy;
  cam->pos.z = cam->target.z + dz;
}

/* Calculates a camera's XYZ position, given that the target
 * and phi + theta + distance are defined */
static void camera_calc_xyz_target( Gtk3dcanvasCameraGl *cam ) {
  float sin_phi, cos_phi, sin_theta, cos_theta;
  float dx, dy, dz;
  
  sin_phi = sin( RAD(cam->phi) );
  cos_phi = cos( RAD(cam->phi) );
  sin_theta = sin( RAD(cam->theta) );
  cos_theta = cos( RAD(cam->theta) );
  
  dx = cam->distance * cos_theta * cos_phi;
  dy = cam->distance * cos_theta * sin_phi;
  dz = cam->distance * sin_theta;
  
  cam->target.x = cam->pos.x - dx;
  cam->target.y = cam->pos.y - dy;
  cam->target.z = cam->pos.z - dz;
}

#if 0 /* FIXME: dead code? */
/* Calculate a target/distance for a camera which only has a location and a
 * phi/theta direction. The target point should be chosen such that it lies on
 * the camera's view ray, and is as close to the origin as possible */
static void camera_make_target( Gtk3dcanvasCameraGl *cam ) {
  float sin_phi, sin_theta;
  float cos_phi, cos_theta;
  float ux, uy, uz;
  float dist;

  sin_phi = sin( RAD(cam->phi) );
  cos_phi = cos( RAD(cam->phi) );
  sin_theta = sin( RAD(cam->theta) );
  cos_theta = cos( RAD(cam->theta) );
  
  /* Calculate unit vector in direction of camera view */
  ux = cos_phi * cos_theta;
  uy = sin_phi * cos_theta;
  uz = sin_theta;

  /* Calculate distance to point on ray nearest the origin */
  dist = cam->pos.x * ux + cam->pos.y * uy + cam->pos.z * uz;

  /* If dist is negative, desired target is behind camera -- doh! */
  cam->distance = MAX(dist, cam->canvas->root->extents.avg / 8.0);
  camera_calc_xyz_target( cam );
}
#endif /* dead code */

/**
 * gtk_3dcanvas_cameragl_mouse_move:
 * @widget: Must be a camera
 * @event: button_press_event, button_release_event or motion_notify_event
 * @nothing: ignored
 * @returns: if the event should be passed further (false in any case)
 *
 * Read mouse input and move camera around accordingly.
 * It is done to be connected to mouse button & motion events 
 */
int gtk_3dcanvas_cameragl_mouse_move( GtkWidget *widget, 
				      GdkEventAny *event, 
				      void *nothing ) {
  static Gtk3dcanvasCameraGl *cam_cur;
  static float mouse_prev_x;
  static float mouse_prev_y;
  static int over_foreign = FALSE;
  GdkEventButton *ev_button;
  GdkEventMotion *ev_motion;
  float mouse_x, mouse_y;
  float mouse_dx, mouse_dy;
  float sin_phi, cos_phi, sin_theta, cos_theta;
  float k;
  int mouse_btn1 = FALSE, mouse_btn2 = FALSE, mouse_btn3 = FALSE;
  int shift_key = FALSE;
  
  switch (event->type) {
  case GDK_BUTTON_PRESS:
    cam_cur = GTK_3DCANVAS_CAMERAGL (widget);
    ev_button = (GdkEventButton *)event;
    /* Note click coordinates (for dragging) */
    mouse_prev_x = ev_button->x;
    mouse_prev_y = ev_button->y;
    /* Change cursor appropriately */
    if (ev_button->button == 1) {
      if (ev_button->state & GDK_SHIFT_MASK)
	set_cursor_glyph( GDK_CIRCLE,widget->window );
      else
	set_cursor_glyph( GDK_FLEUR, widget->window );
    }
    if (ev_button->button == 2)
      set_cursor_glyph( GDK_FLEUR, widget->window );
    if (ev_button->button == 3)
      set_cursor_glyph( GDK_DOUBLE_ARROW, widget->window );
    return FALSE;
    
    /* Double-click -- where are these events coming from? */
  case GDK_2BUTTON_PRESS:
  case GDK_3BUTTON_PRESS:
    return FALSE;
    
  case GDK_BUTTON_RELEASE:
    set_cursor_glyph( GDK_LEFT_PTR, widget->window );
    over_foreign = FALSE;
    return FALSE;
    
  case GDK_MOTION_NOTIFY:
    ev_motion = (GdkEventMotion *)event;
    mouse_x = ev_motion->x;
    mouse_y = ev_motion->y;
    if ((cam_cur != GTK_3DCANVAS_CAMERAGL( widget )) != over_foreign) {
      /* Fix for when pointer is dragged over another GL widget
       * (x/y coords. jump to something else completely) */
      mouse_prev_x = mouse_x;
      mouse_prev_y = mouse_y;
      over_foreign = !over_foreign;
    }
    mouse_dx = cam_cur->mouse_sens * ( mouse_x - mouse_prev_x );
    mouse_dy = cam_cur->mouse_sens * ( mouse_y - mouse_prev_y );
    mouse_prev_x = mouse_x;
    mouse_prev_y = mouse_y;
    /* Get status of Shift key and buttons */
    shift_key = ev_motion->state & GDK_SHIFT_MASK;
    mouse_btn1 = ev_motion->state & GDK_BUTTON1_MASK;
    mouse_btn2 = ev_motion->state & GDK_BUTTON2_MASK;
    mouse_btn3 = ev_motion->state & GDK_BUTTON3_MASK;
    break;
    
  default:
#ifdef DEBUG
    crash( "camera_move( ): invalid message" );
#endif
    return FALSE;
  }
  
  /* Make sure cam->pos is not stale (for Shift + LeftClick) */
  camera_calc_xyz_pos( cam_cur );
  
  /* Button 1: Revolve view around target or camera (Shift for latter) */
  if (mouse_btn1) {
    /* phi = heading, theta = elevation */
    cam_cur->phi -= mouse_dx;
    cam_cur->theta += mouse_dy;
    
    /* Keep angles within proper bounds */
    if (cam_cur->phi < 0.0)
      cam_cur->phi += 360.0;
    if (cam_cur->phi > 360.0)
      cam_cur->phi -= 360.0;
    if (cam_cur->theta < -90.0)
      cam_cur->theta = -90.0;
    if (cam_cur->theta > 90.0)
      cam_cur->theta = 90.0;
  }
  
  /* If Shift was pressed, need to update camera target */
  if (shift_key)
    camera_calc_xyz_target( cam_cur );
  
  sin_phi = sin( RAD(cam_cur->phi) );
  cos_phi = cos( RAD(cam_cur->phi) );
  sin_theta = sin( RAD(cam_cur->theta) );
  cos_theta = cos( RAD(cam_cur->theta) );
  
  /* Button 2: Dolly camera (up/down/left/right) */
  if (mouse_btn2) {
    k = cam_cur->distance / 128;
    /* Isn't trig fun? */
    cam_cur->target.x += k * ( mouse_dx * sin_phi -
			   mouse_dy * sin_theta * cos_phi );
    cam_cur->target.y -= k * ( mouse_dx * cos_phi +
			   mouse_dy * sin_theta * sin_phi );
    cam_cur->target.z += k * mouse_dy * cos_theta;
  }
  
  /* Button 3: Dolly camera (forward/backward) */
  if (mouse_btn3) {
    cam_cur->distance -= (mouse_dy * cam_cur->canvas->root->extents.avg / 48);
    cam_cur->distance = MAX(cam_cur->distance, 
			    cam_cur->canvas->root->extents.avg / 8.0);
  }
  
  /* Now recalculate camera's xyz-position */
  camera_calc_xyz_pos( cam_cur );
  
  /* Real quick, update Camera Position dialog if it is active *
     dialog_Camera_Position( NULL, MESG_(RESET) );
     FIXME: implement it */

  /* Finally, ask for a redraw. */
  gtk_3dcanvas_cameragl_draw(cam_cur);
  
#ifdef DEBUG
  printf("move_camera: phi=%.1f theta=%.1f distance=%.1f pos=%.1f %.1f %.1f\n",
	 cam_cur->phi,cam_cur->theta,cam_cur->distance,
	 cam_cur->target.x,cam_cur->target.y,cam_cur->target.z);
#endif
  return FALSE;
}

/**
 * gtk_3dcanvas_cameragl_expose:
 * @widget: Must be a camera
 * @event: must be expose_event
 * @nothing: ignored
 * @returns: if the event should be passed further
 *
 * Tell the camera something changed. ie, flag it for a redraw.
 * This is done to be connected to the expose_event signal.
 */
int gtk_3dcanvas_cameragl_expose( GtkWidget *widget, 
				    GdkEventExpose *event, 
				    void *nothing ) {
  Gtk3dcanvasCameraGl *cam=GTK_3DCANVAS_CAMERAGL(widget);

  /* draw only last expose */
  if (event->count > 0) {
    return TRUE;
  }

  gtk_3dcanvas_cameragl_draw(cam);
  return FALSE;
}

/**
 * gtk_3dcanvas_cameragl_configure:
 * @widget: Must be a camera
 * @event: must be configure_event
 * @nothing: ignored
 * @returns: if the event should be passed further
 *
 * Resize the camera. Done to be connected to the configure_event signal.
 */
int gtk_3dcanvas_cameragl_configure( GtkWidget *widget, 
				       GdkEventAny *event, 
				       void *nothing ) {
  GdkGLContext *glcontext;
  GdkGLDrawable *gldrawable;
  Gtk3dcanvasCameraGl *cam=GTK_3DCANVAS_CAMERAGL(widget);

  g_return_val_if_fail(widget && event, FALSE);

  glcontext = gtk_widget_get_gl_context(widget);
  gldrawable = gtk_widget_get_gl_drawable(widget);

  if (!glcontext || !gldrawable) {
    printf("glcontext or gldrawable is NULL. Don't configure (yet).\n");
    return FALSE;
  }

  /* OpenGL begin. */
  gdk_gl_drawable_gl_begin(gldrawable, glcontext);

  glViewport (0, 0, widget->allocation.width, widget->allocation.height);

  gdk_gl_drawable_gl_end(gldrawable);
  /* OpenGL end. */

  cam->width  = widget->allocation.width;
  cam->height = widget->allocation.height;
  return TRUE;
}

/**
 * gtk_3dcanvas_cameragl_draw:
 * @cam: The camera to draw
 *
 * Ask for a camera (re)draw.
 */
void gtk_3dcanvas_cameragl_draw( Gtk3dcanvasCameraGl *cam) {
  cam->draw=TRUE;
}

/**
 * gtk_3dcanvas_cameragl_draw:
 * @cam: The camera to draw
 *
 * (Re)draw the camera.
 */
void gtk_3dcanvas_cameragl_do_draw( Gtk3dcanvasCameraGl *cam) {
  
  GdkGLContext *glcontext = gtk_widget_get_gl_context(GTK_WIDGET(cam));
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(GTK_WIDGET(cam));
  float r,g,b;
  float fr_x, fr_y;

  if (!cam->draw) return;
  if (!glcontext || !gldrawable) {
    /* we are called before the widget is ready to draw. */
    return;
  }

  /* OpenGL begin. */
  gdk_gl_drawable_gl_begin(gldrawable, glcontext);
  
  /* fix the background */
  r=cam->background.r;
  g=cam->background.g;
  b=cam->background.b;
  if (cam->dgamma_correct != 1) {
    r = cam->dgamma_lkpt[(int)(r * LKP_TABLE_RES)];
    g = cam->dgamma_lkpt[(int)(g * LKP_TABLE_RES)];
    b = cam->dgamma_lkpt[(int)(b * LKP_TABLE_RES)];
  }
  glClearColor( r, g, b, 0.0 );
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  /* Set view frustum (a.k.a. field of view) */
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity( );
  fr_x = cam->near_clip * tan( RAD(cam->fov) / 2 );
  fr_y = fr_x / ((float)cam->width / (float)cam->height);
  glFrustum( - fr_x, fr_x, - fr_y, fr_y, cam->near_clip, cam->far_clip );

  /* (Re)initialize transformation matrix */
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity( );

  /* Initialize the coordinate system to correct alignment */
  glRotatef( -90.0, 1.0, 0.0, 0.0 );
  glRotatef( -90.0, 0.0, 0.0, 1.0 );
  
  /* Perform world transformations */
  glTranslatef( - cam->distance, 0.0, 0.0 );
  glRotatef( cam->theta, 0.0, 1.0, 0.0 );
  glRotatef( - cam->phi, 0.0, 0.0, 1.0 );
  glTranslatef( - cam->target.x, - cam->target.y, - cam->target.z );
  
  /* For wireframe mode, if active */
  glLineWidth( 2 );
  /* FIXME: the next line pass to wireframe. Shouldn't be hardcoded */
  //glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );

  /* draw world */
#ifdef DEBUG
  printf("Draw it. camera: phi=%.1f theta=%.1f distance=%.1f pos=%.1f %.1f %.1f\n",
	 cam->phi,cam->theta,cam->distance,
	 cam->target.x,cam->target.y,cam->target.z);
  gtk_3dcanvas_group_dump(gtk_3dcanvas_root(cam->canvas),1);
#endif
  draw_group(cam,gtk_3dcanvas_root(cam->canvas));

  /* Draw all active auxiliary objects */
  glDisable( GL_LIGHTING );

  if (cam->axes != 0.0) {
    draw_aux_axes(cam);
  }
  if (cam->grid != 0.0) {
    draw_aux_grid(cam);
  }
  glEnable( GL_LIGHTING );

  /* swap backbuffer to front */
  if (gdk_gl_drawable_is_double_buffered(gldrawable))
    gdk_gl_drawable_swap_buffers(gldrawable);
  else
    glFlush();

  gdk_gl_drawable_gl_end(gldrawable);
  /* OpenGL end. */
  cam->draw = FALSE;
}

static void draw_group(Gtk3dcanvasCameraGl *cam,
		       Gtk3dcanvasGroup *group) {
  int i;

  for (i=0; i<group->childs->len; i++) {
    if (IS_GTK_3DCANVAS_GROUP(g_array_index(group->childs,
					      Gtk3dcanvasGroup*,
					      i))) {
      draw_group(cam,GTK_3DCANVAS_GROUP(g_array_index(group->childs,
							Gtk3dcanvasGroup*,
							i)));
    } else {
      draw_item(cam,GTK_3DCANVAS_ITEM(g_array_index(group->childs,
						      Gtk3dcanvasGroup*,
						      i)));
    }
  }
}

static void draw_item(Gtk3dcanvasCameraGl *cam,
		      Gtk3dcanvasItem *item) {
  /* Execute "before" display list, if there is one */
  if (item->pre_dlist != 0)
    glCallList( item->pre_dlist );
  
  //  printf("draw_item()");
  //  gtk_3dcanvas_item_dump(item);
#ifdef GL_VERSION_1_1
  glInterleavedArrays( GL_C4F_N3F_V3F, 
		       sizeof(Gtk3dOPoint), 
		       item->iarrays->data );
#ifdef GL_VERSION_1_2
  glDrawRangeElements( item->type, 0, 
		       item->iarrays->len - 1, 
		       item->indices->len, 
		       GL_UNSIGNED_INT, 
		       item->indices->data );
#else 
  glDrawElements( item->type, 
		  item->indices->len, 
		  GL_UNSIGNED_INT, 
		  item->indices->data );
#endif /* else GL_VERSION_1_2 */
#else  
  /* Fine, we'll do this the old-fashioned way */
 {
   int i,v;
  glBegin( item->type );
  for (i = 0; i < item->indices->len; i++) {
    v = g_array_index (item->indices, unsigned int, i);
    glColor4fv( &((g_array_index (item->iarrays, Gtk3dOPoint, v)).r) );
    glNormal3fv( &((g_array_index (item->iarrays, Gtk3dOPoint, v)).nx) );
    glVertex3fv( &((g_array_index (item->iarrays, Gtk3dOPoint, v)).x) );
  }
 
  glEnd( );
 }
#endif /* else GL_VERSION_1_1 */
  
  /* Execute "after" display list if there is one */
  if (item->post_dlist != 0)
    glCallList( item->post_dlist );
}

static void draw_aux_axes(Gtk3dcanvasCameraGl *cam) {
}
static void draw_aux_grid(Gtk3dcanvasCameraGl *cam) {

}

/**
 * gtk_3dcanvas_cameragl_set_dgamma_correct:
 * @cam: The camera to change
 * @dgamma: The new value
 *
 * Change the display gamma correction. 1.0 for none, 1.5 for a dark monitor
 * and 0.5 for a bright one.
 */
void gtk_3dcanvas_cameragl_set_dgamma_correct( Gtk3dcanvasCameraGl *cam,
					       float dgamma ) {
  float k, x;
  int i;

  if (dgamma <= 0) { dgamma=0.01; }

  cam->dgamma_correct=dgamma;
  
  k = 1.0 / dgamma;
  for (i = 0; i < LKP_TABLE_RES; i++) {
    x = (float)i / (LKP_TABLE_RES-1);
    cam->dgamma_lkpt[i] = pow( x, k );
  }
}
