/* gtk-3dcanvas-item - stuff we can put on a gtk-3dcanvas */

/* Copyright (C) ????, Daniel Richard G. (base code)    */
/* Copyright (C) 2002, Martin Quinson (objectification) */

/* This program is free software; you can redistribute it and/or modify it
   under the terms of the license (GNU LGPL) which comes with this package. */

#include <stdio.h>
#include <math.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "gtk-3dcanvas.h" /* header of this module */
#include "gtk-3dcanvas-private.h"

/* Useful */
#define PI                      3.14159265358979323846264338327
#define DEG(angleRad)           ((angleRad) * 180 / PI)
#define RAD(angle360)           ((angle360) * PI / 180)
#define SQR(x)                  ((x) * (x))
#define SIGN(x)                 (((x) >= 0) ? 1.0 : -1.0)

/*---------------------------------\
| Property and signals definitions |
\---------------------------------*/
enum {
        ITEM_PROP_0,
        ITEM_PROP_PARENT,
	ITEM_PROP_TYPE,
	ITEM_PROP_COLOR
};

enum {
        ITEM_EVENT,
        ITEM_LAST_SIGNAL
};


/*-----------------\
| Class definition |
\-----------------*/
static void gtk_3dcanvas_item_class_init (Gtk3dcanvasItemClass *class);
static void gtk_3dcanvas_item_init(Gtk3dcanvasItem *glb);

static GObject *
gtk_3dcanvas_item_constructor (GType             type,
			       guint             n_construct_properties,
			  GObjectConstructParam *construct_properties);
static void
gtk_3dcanvas_item_finalize (GObject *object);
static void
gtk_3dcanvas_item_dispose (GObject *object);

static void
gtk_3dcanvas_item_set_property (GObject *object, guint param_id,
                                const GValue *value, GParamSpec *pspec);
static void
gtk_3dcanvas_item_get_property (GObject              *object,
                                guint                 param_id,
                                GValue               *value,
                                GParamSpec           *pspec);


static void
_gtk_3dcanvas_item_update_color(Gtk3dcanvasItem *item);

static gpointer parent_class = NULL;

/*----[ gtk_3dcanvas_item_get_type]-----------------------------------------*/
GType gtk_3dcanvas_item_get_type () {
  static GType type = 0;
  
  if (!type) {
    static const GTypeInfo type_info = {
      sizeof(Gtk3dcanvasItemClass),
      (GBaseInitFunc) NULL,
      (GBaseFinalizeFunc) NULL,
      (GClassInitFunc) gtk_3dcanvas_item_class_init,
      (GClassFinalizeFunc) NULL,
      NULL,                   /* class_data */
      sizeof (Gtk3dcanvasItem),
      0,                      /* n_preallocs */
      (GInstanceInitFunc) gtk_3dcanvas_item_init,
    };

    type = g_type_register_static (GTK_TYPE_OBJECT,
				   "Gtk3dcanvasItem",
				   &type_info, 0);
  }
  
  return type;
}
  

/*----[ gtk_3dcanvas_item_class_init ]--------------------------------------*/
static void gtk_3dcanvas_item_class_init (Gtk3dcanvasItemClass *klass) {
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  object_class->constructor  = gtk_3dcanvas_item_constructor;
  object_class->set_property = gtk_3dcanvas_item_set_property;
  object_class->get_property = gtk_3dcanvas_item_get_property;
  object_class->finalize     = gtk_3dcanvas_item_finalize;
  object_class->dispose      = gtk_3dcanvas_item_dispose;

  g_object_class_install_property
    (object_class, ITEM_PROP_PARENT,
     g_param_spec_object ("parent", 
			  _("Parent"), 
			  _("The parent group in which the object is placed"),
			  GTK_3DCANVAS_GROUP_TYPE,
			  (G_PARAM_READWRITE | G_PARAM_CONSTRUCT)));
  g_object_class_install_property
    (object_class, ITEM_PROP_TYPE,
     g_param_spec_int ("type", 
		       _("Type"),
		       _("OpenGL type (example: GL_QUAD_STRIP)"),
		       0,G_MAXINT,0,
		       (G_PARAM_READWRITE | G_PARAM_CONSTRUCT)));
    
  g_object_class_install_property
    (object_class, ITEM_PROP_COLOR,
     g_param_spec_string ("color", 
			  _("Color"),
			  _("The color, either as name (blue, etc) or hex value in the form '#rgb' '#rrggbb' '#rrrgggbbb' or '#rrrrggggbbbb' where 'r', 'g' and 'b' are hex digits of the red, green, and blue components of the color, respectively"),
			  NULL,
			  (G_PARAM_READWRITE | G_PARAM_CONSTRUCT)));
    

}

/*----[ gl_canvas_item_constructor ]-----------------------------------------*/
static GObject *
gtk_3dcanvas_item_constructor (GType             type,
			       guint            n_construct_properties,
			       GObjectConstructParam *construct_properties)
{
  GObject *object;

  object = G_OBJECT_CLASS (parent_class)->constructor (type,
                                                       n_construct_properties,
                                                       construct_properties);

  return object;
}

/*----[ gtk_3dcanvas_item_finalize ]--------------------------------------*/
static void
gtk_3dcanvas_item_finalize (GObject *object)
{
  G_OBJECT_CLASS (parent_class)->finalize (object);
}

/*----[ gtk_3dcanvas_item_dispose ]---------------------------------------*/
static void
gtk_3dcanvas_item_dispose (GObject *object)
{
  Gtk3dcanvasItem *item = GTK_3DCANVAS_ITEM (object);

  gtk_3dcanvas_group_remove_child(item->group,item);
  G_OBJECT_CLASS (parent_class)->dispose (object);
  /* FIXME: destroy does not work that way, but how ?
  FIXME: should be fixed. Make sure it is.*/
  
}

/*----[ gtk_3dcanvas_item_init ]------------------------------------------*/
static void gtk_3dcanvas_item_init(Gtk3dcanvasItem *item) {
  item->group=NULL;
  item->iarrays = g_array_new(FALSE,FALSE, sizeof(Gtk3dOPoint) );
  item->indices = g_array_new(FALSE,FALSE, sizeof(unsigned int) );
  item->pre_dlist = 0; /* null display list */
  item->post_dlist = 0; /* ditto */
  
  item->extents.xmin = 0;
  item->extents.xmax = 0;
  item->extents.ymin = 0;
  item->extents.ymax = 0;
  item->extents.zmin = 0;
  item->extents.zmax = 0;
  item->extents.avg = 0;
} 

/* Set_property handler for canvas items */
static void
gtk_3dcanvas_item_set_property (GObject *object, guint param_id,
                                const GValue *value, GParamSpec *pspec)
{
  Gtk3dcanvasItem *item;
  PangoColor pango_color;
  
  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_GTK_3DCANVAS_ITEM (object));

  item = GTK_3DCANVAS_ITEM (object);
  
  switch (param_id) {
  case ITEM_PROP_PARENT:
    if (item->group != NULL) {
      g_warning ("Cannot set `parent' argument after item has "
		 "already been constructed.");
    } else if (g_value_get_object (value)) {
      gtk_3dcanvas_group_add_child(GTK_3DCANVAS_GROUP (g_value_get_object (value)),
				   item);
    }
    break;
  case ITEM_PROP_TYPE:
    item->type=g_value_get_int(value);
    break;

  case ITEM_PROP_COLOR:
    if (g_value_get_string (value)) {
      if (pango_color_parse (&pango_color,g_value_get_string (value))) {
	/* FIXME: why do I have to hardcode G_MAXINT? */
	item->color.r = ((double)pango_color.red)/65535.0;
	item->color.g = ((double)pango_color.green)/65535.0;
	item->color.b = ((double)pango_color.blue)/65535.0;
	_gtk_3dcanvas_item_update_color(item);
      } else {
	g_warning("%s is not a color. Cannot change the color of item.",
		  g_value_get_string (value));
      }
    } else {
      /* Passed color is (null) */
      item->color.r = 0;
      item->color.g = 0;
      item->color.b = 0;
      _gtk_3dcanvas_item_update_color(item);
    }
    break;

  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
    break;
  }
}

static void
gtk_3dcanvas_item_get_property (GObject              *object,
                                guint                 param_id,
                                GValue               *value,
                                GParamSpec           *pspec) {
  Gtk3dcanvasItem *item;
  
  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_GTK_3DCANVAS_ITEM (object));
  
  item = GTK_3DCANVAS_ITEM (object);
  
  switch (param_id) {
  case ITEM_PROP_PARENT:
    g_value_set_object (value, item->group);
    break;

  case ITEM_PROP_TYPE:
    g_value_set_int (value, item->type);
    break;
    
  case ITEM_PROP_COLOR:
    g_value_set_string_take_ownership (value,
				       g_strdup_printf ("#%02x%02x%02x",
							(int)(item->color.r * G_MAXINT),
							(int)(item->color.g * G_MAXINT),
							(int)(item->color.b * G_MAXINT)));
    break;

  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
    break;
  }
}

/*---------------------------------\
| non-classical internal functions |
\---------------------------------*/
static void
_gtk_3dcanvas_item_update_color(Gtk3dcanvasItem *item) {
  int i;
  for (i=0;i<item->iarrays->len;i++) {
    g_array_index(item->iarrays,Gtk3dOPoint,i).r=item->color.r;
    g_array_index(item->iarrays,Gtk3dOPoint,i).g=item->color.g;
    g_array_index(item->iarrays,Gtk3dOPoint,i).b=item->color.b;
  }
  gtk_3dcanvas_draw(item->canvas);
}

/*-----------------\
| Public functions |
\-----------------*/

/**
 * gtk_3dcanvas_item_new:
 * @parent: the group in which this item should be placed
 * @type: The object type of the item.
 * @first_arg_name: A list of object argument name/value pairs, 
 * NULL-terminated, used to configure the item. For example, "color", "00BB00",
 * "type", GL_QUAD_STRIP, NULL.
 * @Varargs: Rest of this list
 * @Returns: the newly allocated item
 *
 * Allocates an gtk_3dcanvas item. Once it is created, you have to fill it
 * using add_point and add_index functions.
 *
 * You may find easier to use a child of gtk_3dcanvas_item instead of this.
 *
 **/
Gtk3dcanvasItem *gtk_3dcanvas_item_new(Gtk3dcanvasGroup *parent,
				       GType type,
				       const gchar *first_arg_name, ...) {
  Gtk3dcanvasItem *item;
  va_list args;

  g_return_val_if_fail (IS_GTK_3DCANVAS_GROUP (parent), NULL);
  g_return_val_if_fail (g_type_is_a (type,gtk_3dcanvas_item_get_type ()),NULL);

  //  printf("gtk_3dcanvas_item_new(%p,%ld,%s,...)\n",
  //	 parent,type,first_arg_name);fflush(stdout);

  item = GTK_3DCANVAS_ITEM(g_object_new(type, NULL));

  va_start (args, first_arg_name);
  g_object_set_valist (G_OBJECT (item), first_arg_name, args);
  va_end (args);

  g_object_set(GTK_3DCANVAS_ITEM(item),"parent",parent,NULL);

  return item;
}

/**
 * gtk_3dcanvas_item_add_point:
 * @item: self 
 * @x: x of the point to add
 * @y: y of the point to add
 * @z: z of the point to add
 * @norm_x: the normal along x of the point
 * @norm_y: the normal along y of the point
 * @norm_z: the normal along z of the point
 * @Returns: the index of the created point in the item.
 *
 * Add a point to a previously allocated object.
 *
 * This can be viewed as an internal function for gtk_3dcanvas_item
 *  childs, but was made public just in case you know what you do.
 *
 **/
int gtk_3dcanvas_item_add_point(Gtk3dcanvasItem *item, 
				 float x,      float y,      float z,
				 float norm_x, float norm_y, float norm_z) {

  float d;
  Gtk3dOPoint opoint;
  gboolean extents_changed=FALSE;

  /* Set color */
  opoint.r  = item->color.r;
  opoint.g  = item->color.g;
  opoint.b  = item->color.b;
  /* Initialize "a" fields in iarrays, since we don't really use them */
  opoint.a  = 1.0;

  /* Set vertex location */
  opoint.x = x;
  opoint.y = y;
  opoint.z = z;
  
  /* Normalize and set vertex normal */
  d = sqrt( SQR(norm_x) + SQR(norm_y) + SQR(norm_z) );
  if (d < 1E-6)
    d = 1.0;
  opoint.nx = norm_x / d;
  opoint.ny = norm_y / d;
  opoint.nz = norm_z / d;
  /*
  printf("add_point %d: point: %.2f %.2f %.2f norm: %.2f %.2f %.2f\n",
	 item->iarrays->len,	 
	 opoint.x,opoint.y,opoint.z,
	 opoint.nx,opoint.ny,opoint.nz);
  */
  /* Update the extents */
  if (x < item->extents.xmin || item->iarrays->len == 0) {
    extents_changed = TRUE;
    item->extents.xmin = x;
  }
  if (y < item->extents.ymin || item->iarrays->len == 0) {
    extents_changed = TRUE;
    item->extents.ymin = y;
  }
  if (z < item->extents.zmin || item->iarrays->len == 0) {
    extents_changed = TRUE;
    item->extents.xmin = z;
  }
  if (x > item->extents.xmax || item->iarrays->len == 0) {
    extents_changed = TRUE;
    item->extents.xmax = x;
  }
  if (y > item->extents.ymax || item->iarrays->len == 0) {
    extents_changed = TRUE;
    item->extents.ymax = y;
  }
  if (z > item->extents.zmax || item->iarrays->len == 0) {
    extents_changed = TRUE;
    item->extents.zmax = z;
  }

  /* Store this */  
  g_array_append_val(item->iarrays,opoint);

  /* Recompute the extents if needed */
  if (extents_changed) {
#ifdef DEBUG
    printf("Extents %p of %p changed:x:%.2f-%.2f y:%.2f-%.2f z:%.2f-%.2f avg:%.2f\n",
	   &(item->extents),item,
	   item->extents.xmin,item->extents.xmax,
	   item->extents.ymin,item->extents.ymax,
	   item->extents.zmin,item->extents.zmax,
	   item->extents.avg);
#endif
    item->extents.avg = ((item->extents.xmax - item->extents.xmin) + 
			 (item->extents.ymax - item->extents.ymin) + 
			 (item->extents.zmax - item->extents.zmin)) / 3;
    gtk_3dcanvas_group_compute_extend(item->group);    
  }

  return item->iarrays->len-1;
}

/**
 * gtk_3dcanvas_item_add_index:
 * @item: self
 * @index: the index to add
 *
 * Add an index to a previously allocated item.
 *
 * This can be viewed as an internal function for gtk_3dcanvas_item
 *  childs, but was made public just in case you know what you do.
 **/
void gtk_3dcanvas_item_add_index( Gtk3dcanvasItem *item, int index ) {
  //  printf("gtk_3dcanvas_item_add_index(%d)\n",index);
  g_array_append_val(item->indices,index);
}

/**
 * gtk_3dcanvas_item_empty:
 * @item: self
 *
 * Remove all the previously added points and index from that item.
 *
 * This can be viewed as an internal function for gtk_3dcanvas_item
 *  childs, but was made public just in case you know what you do.
 **/
void gtk_3dcanvas_item_empty( Gtk3dcanvasItem *item ) {
  g_array_set_size(item->iarrays,0);
  g_array_set_size(item->indices,0);
  gtk_3dcanvas_item_compute_extents(item);
}

/**
 * gtk_3dcanvas_item_compute_extents:
 * @item: self
 *
 * Only needed if you remove points from the item, since adding point 
 * updates the extents automatically. But how could you remove points?
 *
 **/
void gtk_3dcanvas_item_compute_extents( Gtk3dcanvasItem *item ) {
  int v;
  float xmin, xmax;
  float ymin, ymax;
  float zmin, zmax;

  if (item->iarrays->len==0) {
    xmin=xmax = ymin=ymax = zmin=zmax = 0;
  } else {
    xmin = xmax = g_array_index(item->iarrays,Gtk3dOPoint,0).x;
    ymin = ymax = g_array_index(item->iarrays,Gtk3dOPoint,0).y;
    zmin = zmax = g_array_index(item->iarrays,Gtk3dOPoint,0).z;

    for (v = 1; v < item->iarrays->len; v++) {
      xmin = MIN(g_array_index(item->iarrays,Gtk3dOPoint,v).x, xmin);
      xmax = MAX(g_array_index(item->iarrays,Gtk3dOPoint,v).x, xmax);
      ymin = MIN(g_array_index(item->iarrays,Gtk3dOPoint,v).y, ymin);
      ymax = MAX(g_array_index(item->iarrays,Gtk3dOPoint,v).y, ymax);
      zmin = MIN(g_array_index(item->iarrays,Gtk3dOPoint,v).z, zmin);
      zmax = MAX(g_array_index(item->iarrays,Gtk3dOPoint,v).z, zmax);
    }
  }
  item->extents.xmin = xmin;
  item->extents.xmax = xmax;
  item->extents.ymin = ymin;
  item->extents.ymax = ymax;
  item->extents.zmin = zmin;
  item->extents.zmax = zmax;
  item->extents.avg = ((xmax - xmin) + (ymax - ymin) + (zmax - zmin)) / 3;

  gtk_3dcanvas_group_compute_extend(item->group);
}

/**
 * gtk_3dcanvas_item_dump
 * @item: what to dump
 *
 * Dump an item to stdout for debugging purpose 
 **/
void gtk_3dcanvas_item_dump(Gtk3dcanvasItem* item) {
  int i;
  if (!item) {
    printf("Item is null\n");
    return;
  }
  printf("Item %p (type %d ; group %p):\n",item,item->type,item->group);
  printf("  extents=x:%.2f-%.2f y:%.2f-%.2f z:%.2f-%.2f avg:%.2f\n",
	 item->extents.xmin,item->extents.xmax,
	 item->extents.ymin,item->extents.ymax,
	 item->extents.zmin,item->extents.zmax,
	 item->extents.avg);
  printf("  color=%.2f %.2f %.2f ; %d iarrays (%p); %d indices (%p)\n",
	 item->color.r,item->color.g,item->color.b,
	 item->iarrays->len,item->iarrays->data,
	 item->indices->len,item->indices->data);
  for (i=0;i<item->iarrays->len;i++) {
    printf("point:%.2f %.2f %.2f; norm:%.2f %.2f %.2f; color:%.2f %.2f %.2f\n",
	   g_array_index(item->iarrays,Gtk3dOPoint,i).x,
	   g_array_index(item->iarrays,Gtk3dOPoint,i).y,
	   g_array_index(item->iarrays,Gtk3dOPoint,i).z,
	   g_array_index(item->iarrays,Gtk3dOPoint,i).nx,
	   g_array_index(item->iarrays,Gtk3dOPoint,i).ny,
	   g_array_index(item->iarrays,Gtk3dOPoint,i).nz,
	   g_array_index(item->iarrays,Gtk3dOPoint,i).r,
	   g_array_index(item->iarrays,Gtk3dOPoint,i).g,
	   g_array_index(item->iarrays,Gtk3dOPoint,i).b);
  }  
  printf("Vertices:");
  for (i=0;i<item->iarrays->len;i++)
    printf(" %d",g_array_index(item->indices,unsigned int,i));
  printf("\n");
}

#if 0
FIXME: Understand, and enable this code
/**
 * gtk_3dcanvas_item_add_stick:
 *
 * Add a ball to the previously allocated item. 
 *
 * FIXME: document the arguments, once I've understood them
 */

void
gtk_3dcanvas_item_add_stick( Gtk3dcanvasItem *item, 
			      float arg1, float arg2, float arg3, float arg4,
			      int num_segs, int smoothness, char alignment,
			      float radius ) {
  float phi;
  float a0, a1, b0, c0;
  float a, b, c;
  float x0, y0, z0;
  float xn, yn, zn;
  float x, y, z;
  int s, p;
  int range_p;
  int reverse = FALSE;
  int end_v, begin_v = 0;
  int obj_v;
  
  /* Interpret args differently depending on alignment */
  switch (alignment) {
  case 'X':
  case 'x':
    a0 = arg1;
    a1 = arg2;
    b0 = arg3;
    c0 = arg4;
    break;
    
  case 'Y':
  case 'y':
    b0 = arg1;
    a0 = - arg2;
    a1 = - arg3;
    c0 = arg4;
    break;
    
  case 'Z':
  case 'z':
    c0 = arg1;
    b0 = arg2;
    a0 = - arg3;
    a1 = - arg4;
    break;
    
  default:
#ifdef DEBUG
    crash( "add_stick( ): invalid alignment" );
#endif
    return;
  }
  
  range_p = 2 * smoothness;
  if (a0 > a1)
    reverse = TRUE; /* to keep the normals right */
  
  /* Endcap vertex (tie point) */
  rotate_xyz( alignment, &x0, &y0, &z0, a0, b0, c0 );
  end_v = add_point( parent_obj, x0, y0, z0, 0.0, 0.0, -1.0 );
  add_index( parent_obj, end_v );
  add_index( parent_obj, end_v );
  
  for (s = 0; s <= num_segs; s++) {
    a = (a1 - a0) * (float)s / (float)num_segs;
    for (p = 0; p < range_p; p++) {
      phi = RAD(360.0 * (float)p / (float)range_p);
      b = STICK_RADIUS * cos( phi );
      c = STICK_RADIUS * sin( phi );
      rotate_xyz( alignment, &x, &y, &z, a, b, c );
      rotate_xyz( alignment, &xn, &yn, &zn, 0, b, c );
      obj_v = add_point( parent_obj, x0 + x, y0 + y, z0 + z, xn, yn, zn );
      if (s > 0) {
	if (reverse) {
	  add_index( parent_obj, obj_v - range_p );
	  add_index( parent_obj, obj_v );
	}
	else {
	  add_index( parent_obj, obj_v );
	  add_index( parent_obj, obj_v - range_p );
	}
      }
      if (p == 0)
	begin_v = obj_v;
    }
    if (s == 0)
      continue;
    
    /* Complete the circle */
    if (reverse) {
      add_index( parent_obj, begin_v - range_p );
      add_index( parent_obj, begin_v );
    }
    else {
      add_index( parent_obj, begin_v );
      add_index( parent_obj, begin_v - range_p );
    }
  }
  
  /* The other endcap vertex (tie point) */
  rotate_xyz( alignment, &x, &y0, &z0, a1, b0, c0 );
  end_v = add_point( parent_obj, x, y0, z0, 0.0, 0.0, -1.0 );
  add_index( parent_obj, end_v );
  add_index( parent_obj, end_v );
}
#endif /* FIXME: enable this code */
