/* gtk-3dcanvas-transfo - simple 3D transformation library */

/* Copyright (C) 2002, Martin Quinson */

/* This program is free software; you can redistribute it and/or modify it
   under the terms of the license (GNU LGPL) which comes with this package. */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "gtk-3dcanvas.h" /* header of this module */
#include "gtk-3dcanvas-private.h"

#ifdef TP
#error someone defined TP, I canot use it in gtk-3dcanvas-tranfo
#endif
#define TP(transfo,a,b) (transfo)[(a)*4+(b)]

/*** PROTOTYPES ***/
inline transfo_t *gtk_3dcanvas_transfo_new(void);
/* c=a*b */
static inline void matmult(transfo_t *a,transfo_t *b,transfo_t *c);

/*** FUNCTIONS ***/
inline void gtk_3dcanvas_transfo_dump(transfo_t *tf) {
  int i,j;
  for (i=0;i<4;i++) {
    for (j=0;j<4;j++) {
      printf("%.3f ",TP(*tf,i,j));
    }
    printf("\n");
  }
}

static inline void matmult(transfo_t *a,transfo_t *b,transfo_t *c) {
  int i,j,k;

  for (i=0;i<4;i++) {
    for (j=0;j<4;j++) {
      TP(*c,i,j)=0;
      for (k=0;k<4;k++) {
	/*	printf("%d,%d += %d,%d(%.0f) * %d,%d(%.0f)\n",
	       i,j,
	       k,j,TP(*a,k,j),
	       i,k,TP(*b,i,k));*/
	TP(*c,i,j) += TP(*a,k,j)*TP(*b,i,k);
      }
    }
  }
}

/**
 * gtk_3dcanvas_transfo_new:
 *
 * Allocate a new transformation
 **/
inline transfo_t *gtk_3dcanvas_transfo_new(void) {
  transfo_t *res=g_malloc(sizeof(transfo_t));
  memset(res,0,sizeof(transfo_t));
  if (res) {
    TP(*res,0,0)=TP(*res,1,1) = 1;
    TP(*res,2,2)=TP(*res,3,3) = 1;
  }
  return res;
}

/**
 * gtk_3dcanvas_transfo_translation:
 * @tf: the transformation
 * @x: x delta
 * @y: y delta
 * @z: z delta
 *
 * Pack a translation to the transformation
 **/
inline void gtk_3dcanvas_transfo_translation(transfo_t *tf,
					     float x,float y,float z) {

  TP(*tf,0,3) += x*TP(*tf,0,0) + y*TP(*tf,0,1) + z*TP(*tf,0,2);
  TP(*tf,1,3) += x*TP(*tf,1,0) + y*TP(*tf,1,1) + z*TP(*tf,1,2);
  TP(*tf,2,3) += x*TP(*tf,2,0) + y*TP(*tf,2,1) + z*TP(*tf,2,2);
  TP(*tf,3,3) += x*TP(*tf,3,0) + y*TP(*tf,3,1) + z*TP(*tf,3,2);

}

/**
 * gtk_3dcanvas_transfo_scale:
 * @tf: the transformation
 * @x: x scaling factor
 * @y: y scaling factor
 * @z: z scaling factor
 *
 * Pack a scale to the transformation
 **/
inline void gtk_3dcanvas_transfo_scale(transfo_t *tf,
				       float x,float y,float z) {

  int i;
  for (i=0;i<4;i++) TP(*tf,i,0) *= x;
  for (i=0;i<4;i++) TP(*tf,i,1) *= y;
  for (i=0;i<4;i++) TP(*tf,i,2) *= z;

}

/**
 * gtk_3dcanvas_transfo_rot_x:
 * @tf: the transformation
 * @theta: rotation angle (in rad)
 *
 * Pack a X-axis rotation to the transformation
 **/
inline void gtk_3dcanvas_transfo_rot_x(transfo_t *tf, float theta) {

  float c=cos(theta),s=sin(theta);
  transfo_t old;
  transfo_t rot={ 1 ,0 ,0 ,0,
		  0 ,c ,-s,0,
		  0 ,s ,c ,0,
		  0 ,0 ,0 ,1 };
  //  printf("roty(%f). c=%f s=%f\n",theta,c,s); 
  memcpy(old,tf,sizeof(transfo_t));

  matmult(&old,&rot,tf);
}

/**
 * gtk_3dcanvas_transfo_rot_y:
 * @tf: the transformation
 * @theta: rotation angle (in rad)
 *
 * Pack a Y-axis rotation to the transformation
 **/
inline void gtk_3dcanvas_transfo_rot_y(transfo_t *tf, float theta) {

  float c=cos(theta),s=sin(theta);
  transfo_t old;
  transfo_t rot={ c ,0,-s,0,
		  0 ,1, 0,0,
		  s ,0, c,0,
		  0 ,0, 0,1 };
  //  printf("roty(%f). c=%f s=%f\n",theta,c,s);
  memcpy(old,tf,sizeof(transfo_t));

  matmult(&old,&rot,tf);
}

/**
 * gtk_3dcanvas_transfo_rot_z:
 * @tf: the transformation
 * @theta: rotation angle (in rad)
 *
 * Pack a Z-axis rotation to the transformation
 **/
inline void gtk_3dcanvas_transfo_rot_z(transfo_t *tf, float theta) {

  float c=cos(theta),s=sin(theta);
  transfo_t old;
  transfo_t rot={ c ,s ,0 ,0,
		  -s,c ,0 ,0,
		  0 ,0 ,1 ,0,
		  0 ,0 ,0 ,1 };
  //  printf("rotz(%f). c=%f s=%f\n",theta,c,s);
  memcpy(old,tf,sizeof(transfo_t));

  matmult(&old,&rot,tf);
}

/**
 * gtk_3dcanvas_transfo_apply
 * @tf: the transformation
 * @x: x of the point to change
 * @y: y of the point to change
 * @z: z of the point to change
 *
 * Apply a transformation to a point
 */

inline void gtk_3dcanvas_transfo_apply(transfo_t *tf, 
				       float *x,float *y,float *z) {
  float oldx=*x,oldy=*y,oldz=*z,h=1;
  *x=TP(*tf,0,0)*oldx + TP(*tf,0,1)*oldy + TP(*tf,0,2)*oldz + TP(*tf,0,3);
  *y=TP(*tf,1,0)*oldx + TP(*tf,1,1)*oldy + TP(*tf,1,2)*oldz + TP(*tf,1,3);
  *z=TP(*tf,2,0)*oldx + TP(*tf,2,1)*oldy + TP(*tf,2,2)*oldz + TP(*tf,2,3);
   h=TP(*tf,3,0)*oldx + TP(*tf,3,1)*oldy + TP(*tf,3,2)*oldz + TP(*tf,3,3);
  if (h==0) h=0.001;
  *x /= h;
  *y /= h;
  *z /= h;
  if (isnan(*x) || isnan(*y) || isnan(*z)) {
    printf("The transformation leaded to NaN!!\n");
    exit(1);
  }
}
